from pathlib import Path
import requests
import requests_mock

import pytest

from sammi.validation import CDFValidator


@pytest.fixture()
def validator():
    """Initialize CDFValidator with default properties."""
    validator = CDFValidator()
    return validator


def test_validate_raw_valid_cdf(validator: CDFValidator):
    """Test validation of raw data."""
    test_data_dir = Path(__file__).parent / "test_data"

    result = validator.validate_raw(test_data_dir / "test_valid.cdf")
    assert type(result) == str
    assert "The following variables are not ISTP-compliant" not in result


def test_validate_raw_invalid_cdf(validator: CDFValidator):
    """Test validation of raw data."""
    test_data_dir = Path(__file__).parent / "test_data"

    result = validator.validate_raw(test_data_dir / "test_invalid.cdf")
    assert type(result) == str
    assert "Global errors" in result
    assert "The following variables are not ISTP-compliant" in result


def test_validate_raw_no_internet(validator: CDFValidator):
    """Test validation when there is no internet connection."""
    test_data_dir = Path(__file__).parent / "test_data"

    with requests_mock.Mocker() as m:
        m.post(validator.api_url, exc=requests.ConnectionError)
        result = validator.validate_raw(test_data_dir / "test_valid.cdf")
        assert "API request failed" in result


def test_validate_raw_server_error(validator: CDFValidator):
    """Test validation when the server returns an error."""
    test_data_dir = Path(__file__).parent / "test_data"

    with requests_mock.Mocker() as m:
        m.post(validator.api_url, status_code=500, text="Internal Server Error")
        result = validator.validate_raw(test_data_dir / "test_valid.cdf")
        assert "API request failed" in result


def test_validate_valid_cdf(validator: CDFValidator):
    """Test validation of a valid CDF file."""
    test_data_dir = Path(__file__).parent / "test_data"

    result = validator.validate(test_data_dir / "test_valid.cdf")
    assert isinstance(result, list)
    assert len(result) == 0


def test_validate_invalid_cdf(validator: CDFValidator):
    """Test validation of an invalid CDF file."""
    test_data_dir = Path(__file__).parent / "test_data"

    result = validator.validate(test_data_dir / "test_invalid.cdf")
    assert isinstance(result, list)
    assert len(result) > 0


def test_validate_no_internet(validator: CDFValidator):
    """Test validation when there is no internet connection."""
    test_data_dir = Path(__file__).parent / "test_data"

    with requests_mock.Mocker() as m:
        m.post(validator.api_url, exc=requests.ConnectionError)
        result = validator.validate(test_data_dir / "test_valid.cdf")
        assert isinstance(result, list)
        assert len(result) == 1
        assert "API request failed" in result[0]


def test_validate_server_error(validator: CDFValidator):
    """Test validation when the server returns an error."""
    test_data_dir = Path(__file__).parent / "test_data"

    with requests_mock.Mocker() as m:
        m.post(validator.api_url, status_code=500, text="Internal Server Error")
        result = validator.validate(test_data_dir / "test_valid.cdf")
        assert isinstance(result, list)
        assert len(result) == 1
        assert "API request failed" in result[0]



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config
import os
import sys
import csv
from pathlib import Path

# The full version, including alpha/beta/rc tags
from sammi import __version__
from sammi.cdf_attribute_manager import CdfAttributeManager

sys.path.insert(0, os.path.abspath(".."))
# -- Project information -----------------------------------------------------

project = "sammi"
copyright = "No rights reserved"
author = "SAMMI Team"

version = __version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "sphinx_copybutton",
]

# Set automodapi to generate files inside the generated directory
automodapi_toctreedirnm = "generated/api"

# -- Generate CSV Files for Docs ---------------------------------------------
if not os.path.exists("generated"):
    os.mkdir("generated")  # generate the directory before putting things in it


def to_csv(info: dict, path: Path):
    # Get Header
    header_set = set()
    for attr_details in info.values():
        header_set.update(attr_details.keys())
    headers = ["Attribute"] + sorted(header_set)

    # Open the output file in write mode
    with open(path, mode="w", newline="", encoding="utf-8") as f:
        writer = csv.writer(f)
        # Write the header row
        writer.writerow(headers)

        # Iterate over the dictionary and write each row
        for attr_name, attr_details in info.items():
            row = [attr_name]
            for key in headers[1:]:  # Skip "Attribute" which is already added
                row.append(attr_details.get(key, ""))
            # Write the row to the CSV
            writer.writerow(row)


# Global Attributes to CSV
global_info = CdfAttributeManager().global_attribute_info()
to_csv(global_info, Path("./generated/global_attributes.csv"))

# Variable Attributes to CSV
variable_info = CdfAttributeManager().variable_attribute_info()
to_csv(variable_info, Path("./generated/variable_attributes.csv"))

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The reST default role (used for this markup: `text`) to use for all
# documents. Set to the "smart" one.
default_role = "obj"

# Disable having a separate return type row
napoleon_use_rtype = False

# Disable google style docstrings
napoleon_google_docstring = False

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": (
        "https://docs.python.org/3/",
        (None, "http://data.astropy.org/intersphinx/python3.inv"),
    ),
    "sunpy": ("https://docs.sunpy.org/en/stable/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "bizstyle"
html_static_path = ["_static"]

html_logo = "logo/sammi_logo.png"
html_favicon = "logo/favicon.ico"
html_css_files = [
    "css/custom.css",
]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]



import pathlib
import requests
from typing import List


class CDFValidator:
    """
    Python class to leverage the Science Physics Data Facility (SPDF)'s API for validation of International Solar-Terrestrial Physics (ISTP) guidelines.

    Parameters
    ----------
    api_url : `str`, optional
        The URL of the SPDF validation API. Default is "https://skteditor.heliophysics.net/cgi-bin/checkcdf.cgi".
    """

    def __init__(
        self,
        api_url: str = "https://skteditor.heliophysics.net/cgi-bin/checkcdf.cgi",
    ):
        self.api_url = api_url

    def validate(self, cdf_path: pathlib.Path) -> List[str]:
        """
        Function to validate a CDF file against the ISTP guidelines using the SPDF validation API.

        Parameters
        ----------
        cdf_path : `pathlib.Path`
            The path to the local CDF file to validate.

        Returns
        -------
        `List[str]`
            A list of error messages from the validation process.
        """
        try:
            raw_response = self.validate_raw(cdf_path)
            return self._parse_errors(raw_response)
        except Exception as e:
            return [f"Validation failed: {str(e)}"]

    def validate_raw(self, cdf_path: pathlib.Path) -> str:
        """
        Function to validate a CDF file against the ISTP guidelines using the SPDF validation API.

        Parameters
        ----------
        cdf_path : `pathlib.Path`
            The path to the local CDF file to validate.

        Returns
        -------
        `str`
            The raw response from the SPDF validation API.
        """
        try:
            with open(cdf_path, "rb") as cdf_to_upload:
                response = requests.post(
                    self.api_url, files={"file": (cdf_path.name, cdf_to_upload)}
                )
                response.raise_for_status()
                return response.content.decode("utf-8")
        except requests.RequestException as e:
            return f"API request failed: {str(e)}"

    def _parse_errors(self, raw_response: str) -> List[str]:
        """
        Parses the raw response from the SPDF validation API to extract error messages.

        Args:
            raw_response (str): The raw string response from the SPDF validation API.

        Returns:
            List[str]: A list of error messages extracted from the raw response.
        """
        errors = []
        current_section = None
        current_variable = None

        if raw_response.startswith("API request failed:"):
            return [raw_response]

        for line in raw_response.splitlines():
            if "Global errors:" in line:
                current_section = "Global errors"
                current_variable = None
            elif "The following variables are not ISTP-compliant" in line:
                current_section = "Variable"
                current_variable = None
            # Two tabs for each new variable name, 4 tabs for each error for that variable.
            elif (
                current_section == "Variable"
                and line.startswith("\t")
                and not line.startswith("\t\t")
            ):
                current_variable = line.strip()
            elif current_section and line.strip() and "Warning" not in line:
                if current_section == "Variable" and current_variable:
                    errors.append(f"{current_variable}:: {line.strip()}")
                else:
                    errors.append(f"{current_section}: {line.strip()}")
            elif current_section and not line.strip():
                current_section = None
                current_variable = None

        # Filter out any entries that are just section headers without actual errors
        errors = [error for error in errors if not error.endswith(":")]

        return errors



from pathlib import Path
import tempfile

import pytest
import yaml

from sammi.cdf_attribute_manager import CdfAttributeManager


@pytest.fixture()
def cdf_manager():
    """Initialize CdfAttributeManager with default properties."""
    cdf_manager = CdfAttributeManager(use_defaults=True)
    return cdf_manager


def test_load_yaml_data():
    """Test Loading Yaml Data for Schema Files"""
    with tempfile.TemporaryDirectory() as tmpdirname:
        # This function writes invalid YAML content into a file
        invalid_yaml = """
        name: John Doe
        age 30
        """

        with open(tmpdirname + "test.yaml", "w") as file:
            file.write(invalid_yaml)

        # Load from an non-existant file
        with pytest.raises(yaml.YAMLError):
            _ = CdfAttributeManager()._load_yaml_data(tmpdirname + "test.yaml")


def test_default_attr_schema(cdf_manager):
    """
    Test function that covers:
        _load_default_global_attr_schema
        _load_default_variable_attr_schema
    """

    # Default global tests
    assert cdf_manager.global_attribute_schema["DOI"]["required"] is False
    assert cdf_manager.global_attribute_schema["Data_type"]["required"] is True

    # Default variable tests
    assert (
        cdf_manager.variable_attribute_schema["attribute_key"]["ABSOLUTE_ERROR"][
            "required"
        ]
        is False
    )
    assert (
        cdf_manager.variable_attribute_schema["attribute_key"]["CATDESC"]["required"]
        is True
    )


def test_cdf_manager_invalid_params():
    """Test Creating a Schema with Invalid Parameters"""
    with pytest.raises(ValueError):
        _ = CdfAttributeManager(
            global_schema_layers=None, variable_schema_layers=None, use_defaults=None
        )


def test_cdf_manager_custom_layers():
    """Test Creating a Schema with Custom Layers"""
    with tempfile.TemporaryDirectory() as tmpdirname:

        # Create Extra Global Layer for Testing
        global_layer_content = """
        test_attribute:
            description: This is a test attribute
            default: null
            required: true
        Data_type:
            required: true   # NOT originally required in Default Schema
        """

        global_test_path = Path(tmpdirname) / "global_test.yaml"
        with open(global_test_path, "w") as file:
            file.write(global_layer_content)
        assert global_test_path.is_file()

        # Create Extra Variable Layer for Testing
        variable_layer_content = """
        attribute_key:
            test_attribute:
                description: This is a test attribute
                required: true
                valid_values: null
                alternate: null
            SI_CONVERSION:
                description: The conversion factor to SI units.
                required: true  # NOT originally required in Default Schema
                valid_values: null
                alternate: null
        data:
            - test_attribute
            - SI_CONVERSION
        support_data:
            - test_attribute
            - SI_CONVERSION
        metadata:
            - test_attribute
        """

        variable_test_path = Path(tmpdirname) / "variable_test.yaml"
        with open(variable_test_path, "w") as file:
            file.write(variable_layer_content)
        assert variable_test_path.is_file()

        cdf_manager = CdfAttributeManager(
            global_schema_layers=[global_test_path],
            variable_schema_layers=[variable_test_path],
            use_defaults=True,
        )

        assert cdf_manager.global_attribute_schema is not None
        # Assert Test Attribute is Added to the Global Schema
        assert "test_attribute" in cdf_manager.global_attribute_schema
        assert cdf_manager.global_attribute_schema["test_attribute"]["required"]
        # Assert Data_type is Overwritten in Global Schema
        assert "Data_type" in cdf_manager.global_attribute_schema
        assert cdf_manager.global_attribute_schema["Data_type"]["required"]
        # Assert other Data_type attributes are not overwritten
        assert (
            cdf_manager.global_attribute_schema["Data_type"]["description"] is not None
        )

        assert cdf_manager.variable_attribute_schema is not None
        # Assert Test Attribute is Added to the Variable Schema
        assert (
            "test_attribute" in cdf_manager.variable_attribute_schema["attribute_key"]
        )
        assert cdf_manager.variable_attribute_schema["attribute_key"]["test_attribute"][
            "required"
        ]
        # Assert SI_CONVERSION is Overwritten in Variable Schema
        assert "SI_CONVERSION" in cdf_manager.variable_attribute_schema["attribute_key"]
        assert cdf_manager.variable_attribute_schema["attribute_key"]["SI_CONVERSION"][
            "required"
        ]

        # Assert Var Type Lists are Updated
        assert len(cdf_manager.variable_attribute_schema["data"]) > 2
        assert len(cdf_manager.variable_attribute_schema["support_data"]) > 2
        assert len(cdf_manager.variable_attribute_schema["metadata"]) > 1
        assert "test_attribute" in cdf_manager.variable_attribute_schema["data"]
        assert "test_attribute" in cdf_manager.variable_attribute_schema["support_data"]
        assert "test_attribute" in cdf_manager.variable_attribute_schema["metadata"]
        assert "SI_CONVERSION" in cdf_manager.variable_attribute_schema["data"]
        assert "SI_CONVERSION" in cdf_manager.variable_attribute_schema["support_data"]
        assert "SI_CONVERSION" not in cdf_manager.variable_attribute_schema["metadata"]


def test_load_global_attribute(cdf_manager):
    """
    Test function that covers:
        load_global_attributes
    """

    test_data_dir = Path(__file__).parent / "test_data"
    # Initialize CdfAttributeManager object which loads in default info
    cdf_manager.load_global_attributes(
        test_data_dir / "imap_default_global_cdf_attrs.yaml"
    )

    # Testing information has been loaded in
    assert cdf_manager._global_attributes["Project"] == "STP>Solar-Terrestrial Physics"
    assert (
        cdf_manager._global_attributes["Source_name"]
        == "IMAP>Interstellar Mapping and Acceleration Probe"
    )
    assert (
        cdf_manager._global_attributes["Discipline"]
        == "Solar Physics>Heliospheric Physics"
    )
    assert (
        cdf_manager._global_attributes["Mission_group"]
        == "IMAP>Interstellar Mapping and Acceleration Probe"
    )
    assert cdf_manager._global_attributes["PI_name"] == "Dr. David J. McComas"
    assert (
        cdf_manager._global_attributes["PI_affiliation"]
        == "Princeton Plasma Physics Laboratory, 100 Stellarator Road, "
        "Princeton, NJ 08540"
    )
    assert (
        cdf_manager._global_attributes["File_naming_convention"]
        == "source_descriptor_datatype_yyyyMMdd_vNNN"
    )
    # The following test should fail because "DOI" is not an attribute in
    #   imap_default_global_cdf_attrs.yaml
    with pytest.raises(KeyError):
        assert cdf_manager._global_attributes["DOI"] == "test"

    # Load in different data
    cdf_manager.load_global_attributes(
        test_data_dir / "default_global_test_cdf_attrs.yaml"
    )

    # Testing attributes carried over
    assert (
        cdf_manager._global_attributes["File_naming_convention"]
        == "source_descriptor_datatype_yyyyMMdd_vNNN"
    )
    assert (
        cdf_manager._global_attributes["Discipline"]
        == "Solar Physics>Heliospheric Physics"
    )

    # Testing attributes newly loaded
    assert cdf_manager._global_attributes["Project"] == "STP>Solar-Terrestrial Physics"
    assert (
        cdf_manager._global_attributes["Source_name"]
        == "IMAP>Interstellar Mapping and Acceleration Probe"
    )
    assert cdf_manager._global_attributes["Mission_group"] == "Dysfunctional Cats"
    assert cdf_manager._global_attributes["PI_name"] == "Ana Manica"
    assert cdf_manager._global_attributes["PI_affiliation"] == "LASP, CU"
    assert cdf_manager._global_attributes["Data_version"] == 2
    assert cdf_manager._global_attributes["DOI"] == "test"

    # Testing that everything loaded into the global attrs is present in
    #   the global attrs schema
    for attr in cdf_manager._global_attributes.keys():
        assert attr in cdf_manager.global_attribute_schema.keys()


def test_get_global_attributes(cdf_manager):
    """
    Test function that covers:
        get_global_attributes
    """
    # Change filepath to load test global attributes
    test_data_dir = Path(__file__).parent / "test_data"
    cdf_manager.load_global_attributes(
        test_data_dir / "default_global_test_cdf_attrs.yaml"
    )
    cdf_manager.load_global_attributes(test_data_dir / "imap_test_global.yaml")

    # Loading in instrument specific attributes
    test_get_global_attrs = cdf_manager.get_global_attributes("imap_test_T1_test")

    # Testing information previously loaded into global attributes (first if statement)
    assert test_get_global_attrs["Project"] == "STP>Solar-Terrestrial Physics"
    assert (
        test_get_global_attrs["Source_name"]
        == "IMAP>Interstellar Mapping and Acceleration Probe"
    )
    assert test_get_global_attrs["Mission_group"] == "Dysfunctional Cats"
    # Testing instrument specific global attributes (first elif statement)
    assert test_get_global_attrs["Descriptor"] == "TEST>Testinstrument"
    assert test_get_global_attrs["Data_type"] == "T1_test-one>Test-1 test one"
    assert test_get_global_attrs["Logical_source"] == "imap_test_T1_test"
    assert (
        test_get_global_attrs["Logical_source_description"]
        == "IMAP Mission TEST one document Level-T1."
    )
    # Not given, and not required information
    with pytest.raises(KeyError):
        assert test_get_global_attrs["bad_name"] == "false info"

    # Testing second elif statement
    test_error_elif = cdf_manager.get_global_attributes("imap_test_T3_test")
    with pytest.raises(KeyError):
        assert test_error_elif["Generation_date"] == "Does Not Exist"

    # Load in more data using get_global_attributes
    test_get_global_attrs_2 = cdf_manager.get_global_attributes("imap_test_T2_test")
    # Testing information previously loaded into global attributes (first if statement)
    assert test_get_global_attrs_2["Project"] == "STP>Solar-Terrestrial Physics"
    # Testing first elif statement
    assert test_get_global_attrs_2["Descriptor"] == "TEST>Testinstrument"
    # "Data_type" not required according to default schema
    assert test_get_global_attrs_2["Data_type"] == "T2_test-two>Test-2 test two"

    # Testing that required schema keys are in get_global_attributes
    for attr_name in cdf_manager.global_attribute_schema.keys():
        required_schema = cdf_manager.global_attribute_schema[attr_name]["required"]
        if required_schema is True:
            assert attr_name in test_get_global_attrs.keys()


def test_instrument_id_format(cdf_manager):
    """
    Test function that covers:
        get_global_attributes
    """
    # Change filepath to load test global attributes
    test_data_dir = Path(__file__).parent / "test_data"
    cdf_manager.load_global_attributes(
        test_data_dir / "imap_default_global_cdf_attrs.yaml"
    )
    cdf_manager.load_global_attributes(test_data_dir / "imap_test_global.yaml")

    # Loading in instrument specific attributes
    test_get_global_attrs = cdf_manager.get_global_attributes("imap_test_T1_test")

    # Testing how instrument_id operates
    assert test_get_global_attrs["Project"] == cdf_manager._global_attributes["Project"]
    assert (
        test_get_global_attrs["Source_name"]
        == cdf_manager._global_attributes["Source_name"]
    )
    assert (
        test_get_global_attrs["Data_type"]
        == cdf_manager._global_attributes["imap_test_T1_test"]["Data_type"]
    )
    assert (
        cdf_manager._global_attributes["imap_test_T1_test"]["Logical_source"]
        == "imap_test_T1_test"
    )
    with pytest.raises(KeyError):
        assert cdf_manager._global_attributes["imap_test_T1_test"]["Project"]


def test_add_global_attribute(cdf_manager):
    """
    Test function that covers:
        add_global_attribute
    """
    # Change filepath to load test global attributes
    test_data_dir = Path(__file__).parent / "test_data"
    cdf_manager.load_global_attributes(
        test_data_dir / "imap_default_global_cdf_attrs.yaml"
    )
    cdf_manager.load_global_attributes(test_data_dir / "imap_test_global.yaml")

    # Changing a dynamic global variable
    cdf_manager.add_global_attribute("Project", "Test Project")
    test_get_global_attrs = cdf_manager.get_global_attributes("imap_test_T1_test")
    assert cdf_manager._global_attributes["Project"] == "Test Project"
    assert test_get_global_attrs["Project"] == "Test Project"

    # Testing adding required global attribute
    cdf_manager._global_attributes.__delitem__("Source_name")
    # Reloading get_global_attributes to pick up deleted Source_name
    test_get_global_attrs = cdf_manager.get_global_attributes("imap_test_T1_test")
    with pytest.raises(KeyError):
        assert cdf_manager._global_attributes["Source_name"]
    assert test_get_global_attrs["Source_name"] is None

    # Adding deleted global attribute
    cdf_manager.add_global_attribute("Source_name", "anas_source")
    assert cdf_manager._global_attributes["Source_name"] == "anas_source"
    # Reloading get_global_attributes to pick up added Source_name
    test_get_global_attrs = cdf_manager.get_global_attributes("imap_test_T1_test")
    assert test_get_global_attrs["Source_name"] == "anas_source"

    # Testing instrument specific attribute
    cdf_manager._global_attributes["imap_test_T1_test"].__delitem__("Logical_source")
    # Reloading get_global_attributes to pick up deleted Source_name
    test_get_global_attrs = cdf_manager.get_global_attributes("imap_test_T1_test")
    with pytest.raises(KeyError):
        assert cdf_manager._global_attributes["imap_test_T1_test"]["Logical_source"]
    assert test_get_global_attrs["Logical_source"] is None


def test_variable_attribute(cdf_manager):
    """
    Test function that covers:
        load_variable_attributes
        get_variable_attributes
    """

    test_data_dir = Path(__file__).parent / "test_data"
    cdf_manager.load_global_attributes(
        test_data_dir / "imap_default_global_cdf_attrs.yaml"
    )
    # Loading in test data
    cdf_manager.load_variable_attributes(test_data_dir / "imap_test_variable.yaml")

    # All variables required to have:
    expected_attributes = [
        "CATDESC",
        "DEPEND_0",
        "DISPLAY_TYPE",
        "FIELDNAM",
        "FILLVAL",
        "FORMAT",
        "LABLAXIS",
        "UNITS",
        "VALIDMIN",
        "VALIDMAX",
        "VAR_TYPE",
    ]

    # Assuring all required attributes are loaded in
    for attr_name in cdf_manager.variable_attribute_schema["attribute_key"]:
        attribute = cdf_manager.variable_attribute_schema["attribute_key"][attr_name]
        if attribute["required"] is True:
            for exp_attr in expected_attributes:
                assert (
                    exp_attr in cdf_manager.variable_attribute_schema["attribute_key"]
                )

    # Testing specific attributes
    assert (
        cdf_manager._variable_attributes["default_attrs"]["DEPEND_0"]
        == cdf_manager._variable_attributes["default_attrs"]["DEPEND_0"]
    )
    assert cdf_manager._variable_attributes["default_attrs"]["FILLVAL"] == -10
    assert cdf_manager._variable_attributes["test_field_1"]["DEPEND_0"] == "test_depend"
    assert (
        cdf_manager._variable_attributes["default_attrs"]["VAR_TYPE"] == "test_var_type"
    )
    with pytest.raises(KeyError):
        assert cdf_manager._variable_attributes["default_attrs"]["CATDESC"] == "test"


def test_get_variable_attributes(cdf_manager):
    """
    Test function that covers:
        load_variable_attributes
        get_variable_attributes
    """
    # Change filepath to load test global attributes
    test_data_dir = Path(__file__).parent / "test_data"
    cdf_manager.load_global_attributes(
        test_data_dir / "imap_default_global_cdf_attrs.yaml"
    )
    # Loading in test data
    cdf_manager.load_variable_attributes(test_data_dir / "imap_test_variable.yaml")

    # Loading in instrument specific attributes
    imap_test_variable = cdf_manager.get_variable_attributes("test_field_1")

    # Make sure all expected attributes are present
    for variable_attrs in cdf_manager.variable_attribute_schema["attribute_key"]:
        required_var_attributes = cdf_manager.variable_attribute_schema[
            "attribute_key"
        ][variable_attrs]["required"]
        if required_var_attributes is True:
            assert variable_attrs in imap_test_variable.keys()

    # Calling default attributes
    assert imap_test_variable["DEPEND_0"] == "test_depend"
    assert imap_test_variable["DISPLAY_TYPE"] == "test_display_type"
    assert imap_test_variable["FILLVAL"] == -10

    # Calling required attributes
    assert imap_test_variable["CATDESC"] == "test time"
    assert imap_test_variable["TIME_BASE"] == 10
    assert imap_test_variable["DEPEND_1"] == "test_depend_1"
    assert imap_test_variable["DEPEND_2"] == "test_depend_2"

    # Calling to non required attributes
    assert imap_test_variable["LEAP_SECONDS_INCLUDED"] == "test_not_required"

    # Calling attribute name that does not exist
    with pytest.raises(KeyError):
        assert imap_test_variable["DOES_NOT_EXIST"]

    # Testing for attribute not in schema
    with pytest.raises(KeyError):
        assert imap_test_variable["NOT_IN_SCHEMA"]

    # Load in different data, test again
    imap_test_variable_2 = cdf_manager.get_variable_attributes("test_field_2")
    # Calling default attributes
    assert imap_test_variable_2["DEPEND_0"] == "test_depend"
    assert imap_test_variable_2["DISPLAY_TYPE"] == "test_display_type"
    assert imap_test_variable_2["FILLVAL"] == -10

    # Calling required attributes
    assert imap_test_variable_2["CATDESC"] == "test time 2"
    assert imap_test_variable_2["TIME_BASE"] == 11

    # Loading in different data to test logger errors, empty strings,
    # and DEPEND_i with i >= 1 condition
    imap_test_variable_3 = cdf_manager.get_variable_attributes("test_field_3")

    assert imap_test_variable_3["DEPEND_1"] == "depend_1_test_3"
    assert imap_test_variable_3["DEPEND_0"] == ""
    assert imap_test_variable_3["CATDESC"] == ""
    assert imap_test_variable_3["REPRESENTATION_2"] == "representation_2"
    assert imap_test_variable_3["LABL_PTR_1"] == "labl_ptr_1"

    # check_schema = False
    imap_test_variable_1_false = cdf_manager.get_variable_attributes(
        "test_field_1", False
    )
    assert imap_test_variable_1_false["NOT_IN_SCHEMA"] == "not_in_schema"
    assert imap_test_variable_1_false["VALIDMIN"] == 0

    var_with_non_ascii_text = cdf_manager.get_variable_attributes("test_field_4")
    assert var_with_non_ascii_text["CATDESC"] == "α-particles"


def test_sw_templates(cdf_manager):
    """Test Global and Variable Attribute Templates"""

    # Global Attribute Template
    assert cdf_manager.global_attribute_template() is not None
    assert isinstance(cdf_manager.global_attribute_template(), dict)

    # Variable Attribute Template
    assert cdf_manager.variable_attribute_template() is not None
    assert isinstance(cdf_manager.variable_attribute_template(), dict)


def test_sw_info(cdf_manager):
    """Test Global and Variable Attribute Info Functions"""

    # Global Attribute Info
    assert cdf_manager.global_attribute_info() is not None
    assert isinstance(cdf_manager.global_attribute_info(), dict)
    assert isinstance(
        cdf_manager.global_attribute_info(attribute_name="Descriptor"), dict
    )
    with pytest.raises(KeyError):
        _ = cdf_manager.global_attribute_info(attribute_name="NotAnAttribute")

    # Variable Attribute Info
    assert cdf_manager.variable_attribute_info() is not None
    assert isinstance(cdf_manager.variable_attribute_info(), dict)
    assert isinstance(
        cdf_manager.variable_attribute_info(attribute_name="CATDESC"), dict
    )
    with pytest.raises(KeyError):
        _ = cdf_manager.variable_attribute_info(attribute_name="NotAnAttribute")
